# built-in modules
import re
import os
import sys
import warnings
import logging
from contextlib import contextmanager

# 3rd party modules
import pandas as pd

logger = logging.getLogger('wvlt.addons')

# Add-ons
def _warning(
    message,
    category = UserWarning,
    filename = '',
    lineno = -1,
    file = None, 
    line = None):
    print("%s: %s" % (category.__name__, message))
warnings.showwarning = _warning

import matplotlib.pyplot as plt
# Reads styles in /styles
stylesheets = plt.style.core.read_style_directory(os.path.join(os.getcwd(), 'style'))
# Update dictionary of styles
plt.style.core.update_nested_dict(plt.style.library, stylesheets)
plt.style.core.available[:] = sorted(plt.style.library.keys())

import pandas as pd
from pandas.api.types import is_numeric_dtype, is_object_dtype
pd.DataFrame.columnstartswith = lambda self, x: [c for c in self.columns if c.startswith(x)]
pd.DataFrame.columnsmatch = lambda self, x: [c for c in self.columns if re.findall(x, c)]
def columnsconditioned(self, start, *args):
    columns = self.columnsmatch(f'^{start}[^_]+$')
    if args:
        for a in args:
            for c in [c for c in columns]:
                if not re.findall(a, c):
                    columns.pop(columns.index(c))
    
    return columns
pd.DataFrame.columnsconditioned = columnsconditioned
def df_to_file(self, file_name, *a, **k): 
    to_functions = {'csv': pd.DataFrame.to_csv,
                    'xlsx': pd.DataFrame.to_excel,
                    'txt': pd.DataFrame.to_csv,
                    'parquet': pd.DataFrame.to_parquet,
                    'temporary': pd.DataFrame.to_parquet,
                    'json': pd.DataFrame.to_json}
    
    file_str = file_name if isinstance(file_name, str) else file_name.name
    if file_str.endswith('.part'):
        file_str = file_str.replace('.part', '')

    for file_ext, to in to_functions.items():
        if file_str.endswith(file_ext):
            return to(self, file_name, *a, **k)
        
    logger.warning(f"File extension not recognized: {file_str}")
    return None
pd.DataFrame.to_file = df_to_file
def pd_read_file(file_name, *a, **k):
    read_functions = {'csv': pd.read_csv,
                    'xlsx': pd.read_excel,
                    'txt': pd.read_csv,
                    'parquet': pd.read_parquet,
                    'temporary': pd.read_parquet,
                    'json': pd.read_json}
    for file_ext, read in read_functions.items():
        if file_name.endswith(file_ext):
            return read(file_name, *a, **k)         
    return None
pd.read_file = pd_read_file

@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:  
            yield
        finally:
            sys.stdout = old_stdout


from . import main, version, _core, _extra, partitioning
from ._core.wavelet_functions import universal_wt as wavelet_transform
# from .main import run_from_eddypro
from ._core.handlers import process, main, run_from_eddypro, integrate_cospectra_from_file, condition_sampling_partition



"""
This script is a key part of the following publications:
    - Herig Coimbra, Pedro Henrique and Loubet, Benjamin and Laurent, Olivier and Mauder, Matthias and Heinesch, Bernard and 
    Bitton, Jonathan and Delpierre, Nicolas and Depuydt, Jérémie and Buysse, Pauline, Improvement of Co2 Flux Quality Through 
    Wavelet-Based Eddy Covariance: A New Method for Partitioning Respiration and Photosynthesis. 
    Available at SSRN: https://ssrn.com/abstract=4642939 or http://dx.doi.org/10.2139/ssrn.4642939

The main function is:  
- run_wt
    function: (1) gets data, (2) performs wavelet transform, (3) cross calculate variables, (4) averages by 30 minutes, (5) saves 
    call: run_wt()
    Input:
        a: 
    Return:
        b: 

- conditional_sampling
    function: split an array (n dimensions) into 4 arrays based on signal (+ or -) of itself and 2nd array 
    call: conditional_sampling()
    Input:
        args: arrays to be used as filter 
    Return:
        b: 

- universal_wt
    function: call any wavelet transform
    call: universal_wt()
    Input:
        a: 
    Return:
        b: 
"""

# built-in modules
import os
import re
import warnings
import logging
import copy
import time
import datetime
import glob

# 3rd party modules
from functools import reduce
import numpy as np
import pandas as pd
import itertools
from scipy.optimize import curve_fit
import pywt
import yaml
try: 
    import pycwt
except ImportError as e:
    pycwt = None
    pass
try: 
    import fcwt
except ImportError as e:
    fcwt = None
    pass

# project modules
from . import commons as hc24
from .read_data import loaddatawithbuffer
from .wavelet_functions import universal_wt, formula_to_vars, prepare_signal, bufferforfrequency_dwt, bufferforfrequency
from ..partitioning.coimbra_et_al_2025 import conditional_sampling, partition_DWCS, partition_DWCS_CO, partition_DWCS_H2O
from .._extra import eddypro_tools as eddypro


logger = logging.getLogger(__name__)



# def sample_raw_data(input_path, datetimerange, acquisition_frequency=20, fileduration=30, **kwargs):
#     raw_kwargs = {'path': input_path, 'fkwargs': {'dt': 1/acquisition_frequency}}
#     kwargs['fmt'] = kwargs.get('fmt', {})
#     if 'gas4_name' in kwargs.keys(): kwargs['fmt'].update({kwargs.pop('gas4_name'): '4th gas'})
#     raw_kwargs.update({k: v for k, v in kwargs.items() if k in ['fmt']})

#     ymd = [datetimerange.split('-')[0], datetimerange.split('-')[1], f'{fileduration}min']
#     _, _, _f = ymd
#     ymd = hc24.list_time_in_period(*ymd, '1D', include='both')

#     for ymd_i, yl in enumerate(ymd):
#         data = hc24.loaddatawithbuffer(
#             yl, d1=None, freq=_f, buffer=0, f_freq=_f, **raw_kwargs)
#         break
#     return data


# def sample_raw_data(input_path, datetimerange, acquisition_frequency=20, fileduration=30, processduration='1D'):
#     ymd = [datetimerange.split(
#         '-')[0], datetimerange.split('-')[1], f'{fileduration}min']
#     _, _, _f = ymd
#     ymd = waveletec._core.list_time_in_period(
#         *ymd, processduration, include='both')

#     for ymd_i, yl in enumerate(ymd):
#         data = waveletec._core.loaddatawithbuffer(
#             yl, d1=None, freq=_f, buffer=0, f_freq=_f, **{'path': input_path, 'fkwargs': {'dt': 1/acquisition_frequency}})
#         break
#     return data

# raw_kwargs = {'path': input_path, 'fkwargs': {'dt': 1/acquisition_frequency}}
# kwargs['fmt'] = kwargs.get('fmt', {})
# if 'gas4_name' in kwargs.keys(): kwargs['fmt'].update({kwargs.pop('gas4_name'): '4th gas'})
# raw_kwargs.update({k: v for k, v in kwargs.items() if k in ['fmt']})

# ymd, raw_kwargs, output_folderpath = None, verbosity = 1,
# overwrite = False, processing_time_duration = "1D",
# internal_averaging = None, dt = 0.05, wt_kwargs = {},
# method = "dwt", averaging = 30, **kwargs)

    

# def eddypro_wavelet_run(site_name, input_path, outputpath, datetimerange, acquisition_frequency=20, fileduration=30, 
#          processduration='1D', integration_period=None, preaverage=None,
#          covariance = None, variables_available=['u', 'v', 'w', 'ts', 'co2', 'h2o'], denoise=0, deadband=[], 
#          method = 'dwt', wave_mother='db6', **kwargs):
#     local_args = locals()

#     if outputpath is not None:
#         hc24.start_logging(outputpath)

#         # Select output file path
#         if method == 'cov':
#             outputpath = str(os.path.join(outputpath, str(site_name)+'{}_{}.csv'))
#         else:
#             outputpath = str(os.path.join(outputpath, 'wavelet_full_cospectra', str(site_name)+'_CDWT{}_{}.csv'))

#         # Save args for run
#         hc24.mkdirs(outputpath)
#         with open(os.path.join(os.path.dirname(os.path.dirname(outputpath)), f'log/setup_{datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")}.yml'), 'w+') as stp:
#             yaml.safe_dump(local_args, stp)

#     # Select covariances
#     # x*y → Cov(x, y)
#     # x*y|x*z|x*... → Cov(x, y)|Cov(x, z),Cov(x, ...)
#     if covariance is None:
#         covariance = hc24.available_combinations(
#             hc24.DEFAULT_COVARIANCE, variables_available)

#     # RUN WAVELET FLUX PROCESSING
#     # ymd = [START_DATE, END_DATE, FILE_FREQUENCY]
#     raw_kwargs = {'path': input_path, 'fkwargs': {'dt': 1/acquisition_frequency}}
#     kwargs['fmt'] = kwargs.get('fmt', {})
#     if 'gas4_name' in kwargs.keys(): kwargs['fmt'].update({kwargs.pop('gas4_name'): '4th gas'})
#     raw_kwargs.update({k: v for k, v in kwargs.items() if k in ['fmt']})
#     data = wavelet_functions.load_data_and_loop(ymd = [datetimerange.split('-')[0], datetimerange.split('-')[1], f'{fileduration}min'],
#                                          output_path = outputpath,
#                                          varstorun = covariance,
#                                          averaging = [fileduration],
#                                          processing_time_duration = processduration,
#                                          method = method,
#                                          wt_kwargs = {'fs': acquisition_frequency, 'wavelet': wave_mother},
#                                          raw_kwargs = raw_kwargs,
#                                          verbosity=5)
#     return data

 

def condition_sampling_partition(folder, output_name=None, 
                                 id_columns=None,
                                 variables_available=['u', 'v', 'w', 'ts', 'co2', 'h2o'], **kwargs):
    # RUN PARTITIONING
    # dst_path = os.path.join(folder, str(
    #     site_name)+f'_CDWT_full_cospectra.csv')
    output_name = output_name or '0000_CDWT_partitioning'
    input_file = glob.glob(os.path.join(folder, '*_full_cospectra*.csv'))

    assert input_file, 'File not found.'
    
    input_file = input_file[0]
    logger.debug(f"pd.read_file('{input_file}')")
    data = pd.read_file(input_file)

    id_columns = id_columns or ['TIMESTAMP'] + [c for c in ['natural_frequency'] if c in data]

    if (len(data.columns) < 5) & ('variable' in data.columns) & ('value' in data.columns):
        data = (data
                .dropna(subset=id_columns)
                .groupby(id_columns + ['variable'])
                .mean()
                .reset_index()
                .pivot(
                    index=id_columns,
                    columns='variable',
                    values='value'
                )
                .reset_index())

    h2o_dw_required_variables = ['w','co2','h2o']
    is_lacking_variable = sum([v not in variables_available for v in h2o_dw_required_variables])
    if not is_lacking_variable:
        logger.debug("partition_DWCS_H2O")
        try:
            partition_DWCS_H2O(
                data, NEE='NEE', GPP='GPP', Reco='Reco', CO2='wco2',
                CO2neg_H2Opos='wco2-wh2o+', 
                CO2neg_H2Oneg='wco2-wh2o-', NIGHT=None)\
                .filter(id_columns + ['NEE', 'GPP', 'Reco'])\
                .to_file(os.path.join(folder, f'{str(output_name)}.CO2_H2O.csv'), index=False)
        except Exception as e:
            logging.warning(str(e))
    else:
        logger.debug(
            f"Missing variables {', '.join([v for v in h2o_dw_required_variables if v not in variables_available])}.")
    
    h2o_co_dw_required_variables = ['w','co2','h2o','co']
    is_lacking_variable = sum([v not in variables_available for v in h2o_co_dw_required_variables])
    if not is_lacking_variable:
        try:
            partition_DWCS_CO(
                data, NEE='NEE', GPP='GPP', Reco='Reco', ffCO2='ffCO2',
                CO2='wco2', 
                CO2neg_H2Opos='wco2-wh2o+', 
                CO2neg_H2Oneg='wco2-wh2o-', 
                CO2pos_COpos='wco2+wco+', 
                CO2pos_COneg='wco2+wco-',
                NIGHT=None)\
                .filter(id_columns + ['NEE', 'GPP', 'Reco', 'ffCO2'])\
                .to_file(os.path.join(folder, f'{str(output_name)}.CO2_H2O_CO.csv'), index=False)
        except Exception as e:
            logging.warning(str(e))
    else:
        logger.debug(
            f"Missing variables {', '.join([v for v in h2o_co_dw_required_variables if v not in variables_available])}.")
    
    co_dw_required_variables = ['w','co2','co']
    is_lacking_variable = sum([v not in variables_available for v in co_dw_required_variables])
    if not is_lacking_variable:
        try:
            partition_DWCS_CO(
                data, NEE='NEE', GPP='GPP', Reco='Reco', ffCO2='ffCO2',
                CO2='wco2', 
                CO2neg_H2Opos=['wco2-wco+', 'wco2-wco-'], 
                CO2neg_H2Oneg=None, 
                CO2pos_COpos='wco2+wco+', 
                CO2pos_COneg='wco2+wco-',
                NIGHT=None)\
                .filter(id_columns + ['NEE', 'GPP', 'Reco', 'ffCO2'])\
                .to_file(os.path.join(folder, f'{str(output_name)}.CO2_CO.csv'), index=False)
        except Exception as e:
            logging.warning(str(e))
    else:
        logger.debug(
            f"Missing variables {', '.join([v for v in co_dw_required_variables if v not in variables_available])}.")
        
    ch4_dw_required_variables = ['w','co2','ch4']
    is_lacking_variable = sum([v not in variables_available for v in ch4_dw_required_variables])
    if not is_lacking_variable:
        try:
            partition_DWCS_CO(
                data, NEE='NEE', GPP='GPP', Reco='Reco', ffCO2='ffCO2',
                CO2='wco2', 
                CO2neg_H2Opos=['wco2-wch4+', 'wco2-wch4-'], 
                CO2neg_H2Oneg=None, 
                CO2pos_COpos='wco2+wch4+', 
                CO2pos_COneg='wco2+wch4-',
                NIGHT=None)\
                .filter(id_columns + ['NEE', 'GPP', 'Reco', 'ffCO2'])\
                .to_file(os.path.join(folder, f'{str(output_name)}.CO2_CH4.csv'), index=False)
        except Exception as e:
            logging.warning(str(e))
    else:
        logger.debug(
            f"Missing variables {', '.join([v for v in ch4_dw_required_variables if v not in variables_available])}.")

def integrate_cospectra(data, f0, dst_path=None):
    data0 = data[(np.isnan(data['natural_frequency']) == False) * (data['natural_frequency'] >= f0)
                 ].groupby(['variable', 'TIMESTAMP'])['value'].agg(np.nansum).reset_index(drop=False)
    data1 = data[np.isnan(data['natural_frequency'])].drop(
        'natural_frequency', axis=1)

    datai = pd.concat([data1[np.isin(
        data1['variable'], data0['variable'].unique()) == False], data0]).drop_duplicates()
    datai = datai.pivot_table('value', 'TIMESTAMP',
                              'variable').reset_index(drop=False)

    if dst_path:
        datai.to_file(dst_path, index=False)
    return datai


def integrate_cospectra_from_file(root, f0, pattern='_full_cospectra_([0-9]+)_', dst_path=None):
    # use glob.glob to find files matching the pattern
    if isinstance(root, str):
        saved_files = {}
        for name in os.listdir(root):
            dateparts = re.findall(pattern, name, flags=re.IGNORECASE)
            if len(dateparts) == 1:
                saved_files[dateparts[0]] = os.path.join(root, name)

        def __read__(date, path):
            r = pd.read_csv(path, skiprows=11, sep=',')
            if 'natural_frequency' not in r.columns: 
                logger.warning(f'Skipping spectral file. Natural frequency column not found ({path}).')
                return pd.DataFrame()
            if r.natural_frequency.dtype != float: print(date, r.natural_frequency.dtype)
            r['TIMESTAMP'] = pd.to_datetime(date, format='%Y%m%d%H%M')
            return r

        data = pd.concat([__read__(k, v) for k, v in saved_files.items()])
    else:
        data = root
    
    return integrate_cospectra(data, f0, dst_path=dst_path)


def decompose_variables(data, variables=['w', 'co2'],
                     nan_tolerance=.3,
                     verbosity=1, identifier='0000', 
                     **kwargs):
    """    Calculate data decomposed with wavelet transform
    """
    # dictionary of wavelet transforms
    φ = {'info_names': []}
    sj = None

    # run by couple of variables (e.g.: co2*w -> mean(co2'w'))
    try:
        logger.debug(f"Debug variables are: {variables} and data columns are: {data.columns.tolist()}")
        assert len(
            variables), 'Empty list of covariances to run. Check available variables and covariances to be performed.'
        for var in variables:
            if var not in φ.keys():
                ready_signal = prepare_signal(
                    data[var], nan_tolerance=nan_tolerance, identifier=identifier)
                # logger.debug(f"signal is ready: {ready_signal.signal.shape}")
                wt_signal = universal_wt(
                    signal=ready_signal.signal, **kwargs, iwt=True)
                # logger.debug(
                #     f"wt_signal is ready: {wt_signal.wave.shape}, {wt_signal.sj}")
                # φ[var], sj
                φ[var] = wt_signal.wave
                φ[f'{var}_qc'] = np.where(ready_signal.signan, 0, wt_signal.coi)
                φ['info_names'] += [var, f'{var}_qc']
                # logger.debug(f"wt_signal is done.")
        φ.update({'sj': wt_signal.sj})
        φ.update({'coi': wt_signal.coi})
    except Exception as e:
        logger.critical(e)
        # logger.error(f"Error in decompose_variables: {e}")
    # return φ
    return type('var_', (object,), φ)


def decompose_data(data, variables=['w', 'co2'], dt=0.05, method='dwt', nan_tolerance=.3, verbosity=1, **kwargs):
    """    Calculate data decomposed with wavelet transform
    """
    assert method in [
        'dwt', 'cwt', 'fcwt'], "Method not found. Available methods are: dwt, cwt, fcwt"
    
    # φ = {}
    # run by couple of variables (e.g.: co2*w -> mean(co2'w'))
    info_t_startvarloop = time.time()
    
    # run wavelet transform
    φ = decompose_variables(data, variables=variables,
                              nan_tolerance=nan_tolerance, **kwargs)
    # sj = φ.pop('sj', None)
    # for v in variables:
    #     if v not in φ.keys():
    #         signal = np.array(data[v])
    #         signan = np.isnan(signal)
    #         N = len(signal)
    #         Nnan = np.sum(signan)
    #         if Nnan:
    #             if (nan_tolerance > 1 and Nnan > nan_tolerance) or (Nnan > nan_tolerance * N):
    #                 logger.warning(
    #                     f"UserWarning: Too much nans ({np.sum(signan)}, {np.round(100*np.sum(signan)/len(signal), 1)}%) in {data.TIMESTAMP.head(-1)[0]}.")
    #         if Nnan and Nnan < N:
    #             signal = np.interp(np.linspace(0, 1, N), 
    #                     np.linspace(0, 1, N)[signan == False],
    #                     signal[signan==False])
    #         φ[v], sj = universal_wt(signal, method, **wt_kwargs, inv=True)
    #         N = len(signal)
        

    logger.debug(f'\t\tDecompose all variables took {round(time.time() - info_t_startvarloop)} s.')
    
    φs_names = []
    # φs_names = [f'valid_{l}' if l else 'valid' for l in φ.sj]
    logger.debug(f'\t\tφ.info_names ({φ.info_names}).')
    for n in φ.info_names:
        if vars(φ)[n].shape[0] > 1:
            for l in φ.sj:  # use '+ ['']' if __integrate_decomposedarrays_in_dictionary__
                if l: φs_names += [f'{n}_{l}'] 
                else: φs_names += [n] 
        else: φs_names += [n]

    # logger.debug(f"\t\tvars(φ).keys(): {vars(φ).keys()}")
    # logger.debug(f'\t\tφs_names: {φs_names} ({len(φs_names)}).')

    # transform 2D arrays to DataFrame
    values = [vars(φ)[n] for n in φ.info_names]
    # logger.debug(f'\t\tTransform 2D arrays to DataFrame with columns `{"`; `".join(φs_names)}`.')
    # logger.debug(f'\t\t{[np.array(v).shape for v in values]}.')

    info_t_mattime = time.time()
    __temp__ = hc24.matrixtotimetable(np.array(data.TIMESTAMP),
                                      np.concatenate(values, axis=0), columns=φs_names)
    logger.debug(
        f'\t\t\tMatrix to timetable took {round(time.time() - info_t_mattime)} s.')
    
    __temp__.set_index('TIMESTAMP', inplace=True)
    # logger.debug(f'\t\tpd.MultiIndex.from_tuples: {[tuple(c.split("_")) for c in __temp__.columns]}.')
    __temp__.columns = pd.MultiIndex.from_tuples([tuple(c.rsplit('_', 1)) for c in __temp__.columns])
    __temp__ = __temp__.stack(1).reset_index(1).rename(columns={"level_1": "natural_frequency"}).reset_index(drop=False)

    #pattern = re.compile(r"^(?P<variable>.+?)_?(?P<natural_frequency>(?<=_)\d+)?$")
    #__temp__ = __temp__.melt(['TIMESTAMP'] + φ.keys())
    #__temp__ = pd.concat([__temp__.pop('variable').str.extract(pattern, expand=True), __temp__], axis=1)
    __temp__['natural_frequency'] = __temp__['natural_frequency'].apply(lambda j: 1/hc24.j2sj(j, 1/dt) if j else np.nan)

    logger.debug(f'\t\tDecompose data (full) took {round(time.time() - info_t_startvarloop)} s.')
    return __temp__


def _calculate_product_from_formula_(data, formula='w*co2|w*h2o'):
    logger = logging.getLogger('wvlt.pipeline._calculate_product_from_formula_')

    φs = {}
    logger.debug(
        f"\t\tformula: {formula} ({type(formula)}).")

    formulavar = formula_to_vars(formula) if isinstance(
        formula, str) else formula
    xy_name = ''.join(formulavar.xy)
    if xy_name not in data.columns:
        for ci, c in enumerate(formulavar.xy):
            XY = XY * np.array(data[c]).conjugate() if ci else data[c]
        φs[xy_name] = XY
        logger.debug(
            f"\t\tDecomposed covariance shape: {XY.shape}, named: {xy_name}.")
    # ({round(XY.shape[1] / (24*60*60/20), 2)} days, for dt=20Hz)
                        
    # logger.debug(f"\t\tformulavar.condsamp_pair: {formulavar.condsamp_pair}.")
    # logger.debug(f"\t\tφs: {φs}.")
    # φ = pd.DataFrame(φs)
    # logger.debug(f"\t\tφ: {φ.head()}.")
    for cs in formulavar.condsamp_pair:
        cs_name = ''.join(cs)
        # logger.debug(f"\t\tcs_name: {cs_name} (from {cs}) | φs.keys(): {φs.keys()}.")
        if (cs_name not in φs.keys()) and (cs_name not in data.columns):
            for ci, c in enumerate(cs):
                # logger.debug(f"\t\tCurrent c: {c}.")
                # logger.debug(f"\t\tCurrent data: {data.head()}.")
                # logger.debug(f"\t\tDecomposed covariance shape: {data[c].shape}.")
                CS = CS * np.array(data[c]).conjugate() if ci else data[c]
            φs[cs_name] = CS

    # data = pd.concat([data, pd.DataFrame(φs)], axis=1)
    return pd.DataFrame(φs)


def _calculate_conditional_sampling_from_formula_(data, formula='w*co2|w*h2o'):
    logger = logging.getLogger('wvlt.pipeline._calculate_conditional_sampling_from_formula_')

    logger.debug(f"\t\tformula: {formula}.")
    formulavar = formula_to_vars(formula) if isinstance(formula, str) else formula
    
    # logger.debug(f"\t\tformulavar: {formulavar}.")
    names = [''.join(formulavar.xy)] + [''.join(cs)
                                        for cs in formulavar.condsamp_pair]

    logger.debug(f"\t\tnames: {names} ({data.columns.to_list()}).")

    φc = conditional_sampling(
        np.array(data[names[0]]), *[np.array(data[n]) for n in names], names=names, label={1: "+", -1: "-"}) if names else {}

    # data = pd.concat([data, pd.DataFrame(φs)], axis=1)
    return pd.DataFrame(φc)


def __save_cospectra__(data, dst_path, overwrite=False, **meta):
    logger = logging.getLogger('wvlt.pipeline.__save_cospectra__')

    # saved_files = []

    info_t_startsaveloop = time.time()

    # for __datea__, __tempa__ in data.groupby(data.TIMESTAMP):
    # dst_path = output_path.format(pd.to_datetime(__datea__).strftime('%Y%m%d%H%M'))
    logger.debug(f'\t\tSaving {dst_path} with shape {data.shape}.')
    # if os.path.exists(dst_path): continue
    use_header = False

    if overwrite or (not os.path.exists(dst_path)):
        use_header = True
        header  = "wavelet_based_(co)spectra\n"
        header += f"--------------------------------------------------------------\n"
        header += f"TIMESTAMP_START = {meta.get('TIMESTAMP_START', min(data.TIMESTAMP))}\n"
        header += f"TIMESTAMP_END = {meta.get('TIMESTAMP_END', max(data.TIMESTAMP))}\n"
        header += f"N: {meta.get('N', len(data.TIMESTAMP))}\n"
        header += f"TIME_BUFFER [min] = {meta.get('buffer', np.nan)/60}\n"
        header += f"frequency [Hz]\n"
        header += f"y-axis -> wavelet_reconstructed\n"
        header += f"mother_wavelet -> {meta.get('method', '')}\n"
        header += f"acquisition_frequency [Hz] = {1/meta.get('dt', np.nan)}\n"
        header += f"averaging_interval [Min] = {meta.get('averaging', '')}\n"
        hc24.mkdirs(dst_path)
        with open(dst_path, 'w+') as part: part.write(header)
        # legitimate_to_write = 1
        logger.debug(f'\t\tSaving header of DataFrame took {round(time.time() - info_t_startsaveloop)} s.')
        # saved_files.append(dst_path)
    
    # if not legitimate_to_write: continue
    
    data.drop('TIMESTAMP', axis=1, inplace=True)
    with open(dst_path, 'a+', newline='') as part:
        data.to_file(part, header=use_header, chunksize=500, index=False)
    logger.debug(f'\t\tSaving DataFrame took {round(time.time() - info_t_startsaveloop)} s.')
    
    # del data
        
    #arr_slice = np.unique(data.TIMESTAMP, return_index=True)
    #for __datea__ in arr_slice[0]:
    #    dst_path = output_path.format(suffix, pd.to_datetime(__datea__).strftime('%Y%m%d%H%M'))
    #    if os.path.exists(dst_path+'.part'): os.rename(dst_path+'.part', dst_path)
    
    # return saved_files
    return

def process(#ymd, raw_kwargs, 
            datetimerange, fileduration, input_path, acquisition_frequency,
            covariance=None, output_folderpath=None, verbosity=1,
            overwrite=False, processing_time_duration="1D", 
            internal_averaging=None, dt=0.05, wt_kwargs={}, 
            integration_period=None,
            method="dwt", averaging=30, meta={}, **kwargs):
    logger.debug('--- Starting process ---')
    local_args = locals()
    info_t_start = time.time()

    def _date_from_yl(date):
        date = re.sub('[-: ]', '', str(date))
        if processing_time_duration.endswith("D"):
            date = date[:8]
        if processing_time_duration.endswith("H") or processing_time_duration.endswith("Min"):
            date = date[:12]
        return date
    
    def _validate_run(date, yl, compare_start=True, compare_end=False):
        # recheck if files exist and overwrite option
        # doesn't save time (maybe only save 5min)
        file_name = os.path.basename(output_pathmodel.format(date))
        part_name0 = file_name.rsplit('_', 1)[0] + '_' if compare_start else ''
        part_name1 = file_name.rsplit('.', 1)[-1] if compare_end else ''
        current_files = [p for p in os.listdir(os.path.dirname(
            output_pathmodel)) if p.startswith(part_name0) and p.endswith(part_name1)]

        if not overwrite and current_files: #file_name in os.path.exists(output_pathmodel.format(date)):
            logger.warning(
                "UserWarning: Skipping, file already exists ({}).".format(date))
            return False

        # # if all([os.path.exists(output_pathmodel.format(_yl.strftime('%Y%m%d%H%M'))) for _yl in yl[:-1]]):
        # if all([output_pathmodel.format(_yl.strftime('%Y%m%d%H%M')) in current_files for _yl in yl[:-1]]):
        #     logger.warning(
        #         "UserWarning: Skipping, file already exists ({}).".format(date))
        #     return False
        # # elif any([os.path.exists(output_pathmodel.format(_yl.strftime('%Y%m%d%H%M'))) for _yl in yl[:-1]]):
        # if any([output_pathmodel.format(_yl.strftime('%Y%m%d%H%M')) in current_files for _yl in yl[:-1]]):
        #     logger.warning("UserWarning: Continuing but some files already exist ({}), others don't ({}).".format(
        #         ', '.join([_yl.strftime('%Y%m%d%H%M') for _yl in yl[:-1] if output_pathmodel.format(_yl.strftime('%Y%m%d%H%M')) in current_files]),
        #         ', '.join([_yl.strftime('%Y%m%d%H%M') for _yl in yl[:-1] if not output_pathmodel.format(_yl.strftime('%Y%m%d%H%M')) in current_files]),
        #     ))
        #     return True

        if hc24.checkifinprogress(curoutpath_inprog):
            return False
        return True
        
    def _load_data():
        start_time = time.time()
        data = loaddatawithbuffer(yl, d1=None, freq=_f, buffer=buffer, f_freq=_f, **load_kwargs)
        if data.empty:
            if verbosity > 1:
                logger.warning(f"UserWarning: No file found ({date}, path: {load_kwargs.get('path', 'default')}).")
            return None
        logger.debug(f'\tLoading data took {round(time.time() - start_time)} s.')
        return data
    
    def _exit():
        if os.path.exists(curoutpath_inprog):
            os.remove(curoutpath_inprog)

    # Group parameters for each function
    load_kwargs = {
        # 'datetimerange': datetimerange,
        # 'fileduration': fileduration,
        'path': input_path,
        # 'acquisition_frequency': acquisition_frequency,
        'fkwargs': {'dt': 1/acquisition_frequency},
        'fmt': kwargs.get("fmt", {}),  # allow user to override or extend
        **kwargs.get("load_kwargs", {}),  # allow user to override or extend,
    }

    # kwargs['fmt'] = kwargs.get('fmt', {})
    if 'gas4_name' in kwargs.keys():
        load_kwargs['fmt'].update({kwargs.pop('gas4_name'): '4th gas'})

    # raw_kwargs = {'path': input_path, 'fkwargs': {
    #     'dt': 1/acquisition_frequency}}
    # raw_kwargs.update({k: v for k, v in kwargs.items() if k in ['fmt']})

    transform_kwargs = {
        'dt': 1/acquisition_frequency,
        'method': method,
        # 'averaging': averaging,
        'varstorun': covariance or hc24.available_combinations(hc24.DEFAULT_COVARIANCE),
        **kwargs.get("transform_kwargs", {}),
        **wt_kwargs
    }

    output_kwargs = {
        'output_folderpath': output_folderpath,
        'overwrite': overwrite,
        'meta': {'acquisition_frequency': acquisition_frequency},
        **kwargs.get("output_kwargs", {})
    }
    
    logging_kwargs = {
        **kwargs.get("logging_kwargs", {})
    }

    general_config = {
        'sitename': kwargs.get('sitename', '00000'),
        'verbosity': verbosity,
        'processing_time_duration': processing_time_duration,
    }

    
    run_time = datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")
    output_path = ""
    output_pathmodel = ""
    curoutpath_inprog = ""

    # method = kwargs.get('method', 'dwt')
    # sitename = kwargs.get('sitename', '00000')

    # if covariance is None:
    #     # TODO: get variables available from eddypro
    #     covariance = hc24.available_combinations(hc24.DEFAULT_COVARIANCE)

    # kwargs.update({'covariance': covariance,
    #           'processduration': processduration, })
    
    if output_folderpath is not None:
        hc24.start_logging(output_folderpath, **logging_kwargs)
        if method == 'cov':
            output_pathmodel = str(os.path.join(output_folderpath, str(
                general_config['sitename'])+'_cov_{}_'+run_time+'.csv'))
        else:
            output_pathmodel = str(os.path.join(output_folderpath, 'wavelet_full_cospectra', str(
                general_config['sitename'])+'_full_cospectra_{}_'+run_time+'.csv'))
        hc24.mkdirs(output_pathmodel)
        try:
            hc24.save_locals(local_args, os.path.join(output_folderpath, f'log/setup_{run_time}.yml'))
        except Exception as e:
            logger.warning(f"Could not save local arguments to file: {e}")
        # with open(, 'w+') as stp:
        #     yaml.safe_dump(local_args, stp)
    else:
        output_pathmodel = ""
    
    logger.debug(f'Output path: {output_pathmodel}.')
    
    ymd = [datetimerange.split(
        '-')[0], datetimerange.split('-')[1], f'{fileduration}min']
    
    # if isinstance(averaging, (list, tuple)):
    #     averaging = averaging[-1]
    # if internal_averaging is None:
    #     internal_averaging = averaging
    
    fulldata = pd.DataFrame()
    
    _, _, _f = ymd
    ymd = hc24.list_time_in_period(*ymd, processing_time_duration, include='both')

    buffer = 0 if method == 'cov' else (
        bufferforfrequency_dwt(n_=_f, **wt_kwargs) / 2 if method == 'dwt'
        else bufferforfrequency(wt_kwargs.get("f0", 1/(3*60*60))) / 2
    )
    meta.update({'buffer': buffer})
    # logger.debug(f"Buffer (s): {buffer}.")
    logger.debug(
        f'Start date loop at {round(time.time() - info_t_start)} s.')

    # Skip two line
    prev_print = '\n'
    for yl in ymd:
        info_t_yl_ymd = time.time()
        date = _date_from_yl(yl[0])

        print(prev_print, date, 'reading', ' '*10, sep=' ', end='\n')
        prev_print = '\x1B[1A\r'

        if output_pathmodel:
            curoutpath_inprog = output_pathmodel.format(date).rsplit(".", 1)[
                0] + ".inprogress"
            logger.debug(f'In progress file: {curoutpath_inprog}.')
            if not _validate_run(date, yl):
                continue
        
        data = _load_data()
        if data is None:
            _exit()
            continue
        

        try:
            # main run
            output_path = output_pathmodel.format("{}")
            # # run directly
            # output_kwargs.update({'output_path': output_path})
            # fulldata = main(data, period=[min(yl), max(yl)],
            #                 output_kwargs=output_kwargs, **transform_kwargs)
            
            # run by varstorun
            output_kwargs.update({'output_path': output_path + '.part'})
            allvars = transform_kwargs['varstorun']
            saved_files = []
            for f in allvars:
                transform_kwargs['varstorun'] = [f]
                output = main(data, period=[min(yl), max(yl)], 
                              meta=meta,
                              output_kwargs=output_kwargs, **transform_kwargs)
                saved_files.append(output.saved)
                fulldata = pd.concat([fulldata, output.data], axis=0)
                
            for f in [s for s_ in saved_files for s in s_]:
                if os.path.exists(f):
                    os.rename(f, f.replace('.part', ''))
        except Exception as e:
            logger.critical(e)
            print(str(e))

        logger.debug(
            f'Date loop ({yl}) took {round(time.time() - info_t_yl_ymd)} s.')
        _exit()
    
    logger.debug(f'End date loop at {round(time.time() - info_t_start)} s.')
    logger.debug(f"integration_period: {integration_period}.")
    logger.debug(f"output_pathmodel: {output_pathmodel}.")
    logger.debug(f"fulldata: {fulldata.head()}.")
    if output_pathmodel and not fulldata.empty:
        # timestamp = pd.Timestamp.now().strftime('%Y%m%dT%H%M%S_%f')
        dst_path = os.path.join(output_folderpath, os.path.basename(
            output_pathmodel.format(run_time)))
        if integration_period:
            fulldata = integrate_cospectra(fulldata, 1/integration_period, dst_path=None)
        fulldata.to_csv(dst_path, index=False)

    logger.debug(
        f'\t\tFull process took {round(time.time() - info_t_start)} s (run_wt).')
    return fulldata


def main(data, varstorun, period=None, average_period='30min', output_kwargs={}, meta={}, **kwargs):
    info_t_main = time.time()
    vars_unique = list(set([var for f in varstorun for var in formula_to_vars(f).uniquevars]))

    logger.debug(
        f'Input data is ready, data shape is {data.shape}, with unique vars: {vars_unique}, based on {"; ".join(varstorun)}.')
    # logger.debug(
    #     f'\n{data.head()}\n\n{min(data["TIMESTAMP"])} - {max(data["TIMESTAMP"])}\n\n{period}')

    # decompose all required variables
    wvvar = decompose_data(data, vars_unique,
                           nan_tolerance=.3,
                           verbosity=1, identifier='0000',
                           **kwargs)
    meta.update({'averaging': average_period,
                 'method': f"{kwargs.get('method', '')} ~{kwargs.get('mother_wavelet', '')}",
                 'dt': kwargs.get('dt', np.nan)})
    
    # logger.debug(f'Decompose data is over, data shape is {wvvar.shape}.')
    # logger.debug(f'\n{wvvar.head()}\n')
    
    # select valid dates
    if period: wvvar = wvvar[(wvvar['TIMESTAMP'] > period[0]) & (wvvar['TIMESTAMP'] < period[1])]
    wvvar = wvvar.reset_index(drop=True)

    # logger.debug(f'Screen data over period of interest yielded data shape {wvvar.shape}.')

    # calculate covariance
    info_t_calc_product = time.time()
    # wvout = _calculate_product_from_formula_(wvvar, varstorun)
    # logger.debug(f'varstorun. {varstorun}')
    uniquecovs = list(set(
        [c for f in varstorun for c in formula_to_vars(f).combinations]))
    # logger.debug(f'uniquecovs. {uniquecovs}')

    wvout = pd.concat(
        # [wvvar[['TIMESTAMP', 'natural_frequency']]] +
        [(_calculate_product_from_formula_(wvvar, formula=f)
        #   .drop(columns=formula_to_vars(f).uniquevars if i == 0 else ['TIMESTAMP'] + formula_to_vars(f).uniquevars)
          )
         for i, f in enumerate(uniquecovs)], axis=1)
    logger.debug(
        f'\tCalculate product from formula took {round(time.time() - info_t_calc_product)} s.')

    growingdata = pd.concat([wvvar, wvout], axis=1)
    # logger.debug(f'Growing data shape {growingdata.shape}.')

    # calculate conditional sampling    
    info_t_calc_cond_samp = time.time()
    wvcsp = pd.concat(
        # [wvvar[['TIMESTAMP', 'natural_frequency']]] +
        [_calculate_conditional_sampling_from_formula_(growingdata, f)
         for f in varstorun], axis=1)
    logger.debug(
        f'\tCalculate conditional sampling took {round(time.time() - info_t_calc_cond_samp)} s.')
         
    # logger.debug(f'Calculate _calculate_conditional_sampling_from_formula_ is over, with data: {wvcsp.head()}.')

    # despike
    # denoise
    # smoothing
    # Y12 = smooth_2d_data(Y12, method='repeat', smoothing=smoothing)
    # for i in range(len(φcs)):
    #     φcs[i] = smooth_2d_data(
    #         φcs[i], method='repeat', smoothing=smoothing)

    # assemble data
    info_t_assemble_data = time.time()
    growingdata = pd.concat([growingdata, wvcsp], axis=1)
    logger.debug(
        f'\tAssemble data took {round(time.time() - info_t_assemble_data)} s.')

    # average
    info_t_average = time.time()
    for thisdate, thisdata in growingdata.groupby(growingdata['TIMESTAMP'].dt.floor(
            average_period)):
        thisdate_ = thisdate.strftime('%Y%m%d%H%M')
        meta.update({thisdate_: meta})
        meta[thisdate_].update({
                    'TIMESTAMP_START': min(thisdata['TIMESTAMP']),
                    'TIMESTAMP_END': max(thisdata['TIMESTAMP']),
                    'N': len(thisdata['TIMESTAMP'])})

    growingdata['TIMESTAMP'] = growingdata['TIMESTAMP'].dt.floor(
        average_period)
    __ID_COLS__ = list(
        set(['TIMESTAMP', 'natural_frequency']) & set(growingdata.columns))
    growingdata = growingdata.groupby(__ID_COLS__).agg(
        np.nanmean).reset_index(drop=False)
    
    logger.debug(
        f'\tAveraging data took {round(time.time() - info_t_average)} s.')
    
    # integrate

    # save in dataframe and .csv
    growingdata = (growingdata.sort_values(by=__ID_COLS__)
                .melt(__ID_COLS__))
    
    saved_files = []
    if output_kwargs.get('output_path', None):
        logger.debug(f"\tSaving data in {output_kwargs['output_path']}.")
        info_t_save_cospectra = time.time()
        for thisdate, thisdata in growingdata.groupby(growingdata.TIMESTAMP):
            thisdate_ = thisdate.strftime('%Y%m%d%H%M')
            dst_path = output_kwargs.get('output_path').format(thisdate_)
            logger.debug(f"\t\t\t... in {dst_path}.")
            __save_cospectra__(thisdata, dst_path, **meta[thisdate_])
            saved_files += [dst_path]
        logger.debug(
            f'\tSaving data took {round(time.time() - info_t_save_cospectra)} s.')
    
    # rename saved files when done
    # os.rename(output_kwargs['output_path'].format(suffix, pd.datetime.now().strftime('%Y%m%dT%H%M%S_%f')),
    #           )

    logger.debug(
        f'\tMain took {round(time.time() - info_t_main)} s.')
    # save in .nc
    return type('var_', (object,), {'data': growingdata, 'saved': saved_files})


def run_from_eddypro(path,
                     # ="input/EP/FR-Gri_sample.eddypro",
                     #  covariance=["w*co2|w|co2|h2o", "w*co2|w*h2o", "w*h2o",],
                     #  processduration='6H',
                     **kwargs):
    c = eddypro.extract_info_from_eddypro_setup(eddypro=path)
    c.update(**kwargs)

    for path in ['input_path', 'output_folderpath']:
        if c.get(path, None) is not None:
            c[path] = os.path.abspath(c[path])

    return process(**c)



"""
This script is a key part of the following publications:
    - Herig Coimbra, Pedro Henrique and Loubet, Benjamin and Laurent, Olivier and Mauder, Matthias and Heinesch, Bernard and 
    Bitton, Jonathan and Delpierre, Nicolas and Depuydt, Jérémie and Buysse, Pauline, Improvement of Co2 Flux Quality Through 
    Wavelet-Based Eddy Covariance: A New Method for Partitioning Respiration and Photosynthesis. 
    Available at SSRN: https://ssrn.com/abstract=4642939 or http://dx.doi.org/10.2139/ssrn.4642939

The main function is:  
- run_wt
    function: (1) gets data, (2) performs wavelet transform, (3) cross calculate variables, (4) averages by 30 minutes, (5) saves 
    call: run_wt()
    Input:
        a: 
    Return:
        b: 

- conditional_sampling
    function: split an array (n dimensions) into 4 arrays based on signal (+ or -) of itself and 2nd array 
    call: conditional_sampling()
    Input:
        args: arrays to be used as filter 
    Return:
        b: 

- universal_wt
    function: call any wavelet transform
    call: universal_wt()
    Input:
        a: 
    Return:
        b: 
"""

# built-in modules
import os
import re
import warnings
import logging
import copy
import time

# 3rd party modules
from functools import reduce
import numpy as np
import pandas as pd
import pywt
try: 
    import pycwt
except ImportError as e:
    pycwt = None
    pass
try: 
    import fcwt
except ImportError as e:
    fcwt = None
    pass

# project modules

logger = logging.getLogger(__name__)


def __wavemother_str_pycwt__(name):
    wavelets = {w.lower(): vars(pycwt.mothers)[w] for w in ['Morlet', 'Paul', 'DOG', 'MexicanHat']}
    mother = wavelets[re.subn('[0-9]', '',  name.lower())[0]]
    if re.findall('[0-9]+', name.lower()): 
        mother = mother(int(re.findall('[0-9]+', name.lower())[0]))
    else:
        mother = mother()
    return mother


def bufferforfrequency_dwt(N=0, n_=None, fs=20, level=None, f0=None, max_iteration=10**4, wavelet='db6'):
    try:
        import pywt
    except Exception as e:
        logger.error(f"Error in bufferforfrequency_dwt:\n{e}")
    if level is None and f0 is None: f0 = 1/(2*60*60)  # 18
    lvl = level if level is not None else int(np.ceil(np.log2(fs/f0)))
    if n_ is None: n_ = fs * 60 * 30
    n0 = N
    cur_iteration = 0
    while True:
        n0 += pd.to_timedelta(n_)/pd.to_timedelta("1s") * fs if isinstance(n_, str) else n_
        if lvl <= pywt.dwt_max_level(n0, wavelet):
            break
        cur_iteration += 1
        if cur_iteration > max_iteration:
            warnings.warn('Limit of iterations attained before buffer found. Current buffer allows up to {} levels.'.format(
                pywt.dwt_max_level(n0, wavelet)))
            break
    return (n0-N) * fs**-1


def bufferforfrequency(f0, dt=0.05, param=6, mother="MORLET", wavelet=None):
    try:
        import pywt
    except Exception as e:
        logger.error(f"Error in bufferforfrequency:\n{e}")
    wavelet = wavelet or pycwt.Morlet(6)
    #check if f0 in right units
    # f0 ↴
    #    /\
    #   /  \
    #  /____\
    # 2 x buffer
    
    if isinstance(wavelet, str): wavelet = __wavemother_str_pycwt__(wavelet)
    c = wavelet.flambda() * wavelet.coi()
    n0 = 1 + (2 * (1/f0) * (c * dt)**-1)
    N = int(np.ceil(n0 * dt))
    return N


def formula_to_vars(formula):
    """
    function: parse formula to variables
    call: formula_to_vars()
    Input:
        formula: string
    Return:
        var_: object with attributes xy, condsamp_pair, uniquevars
    """
    # parse formula
    xy = formula.split('|')[0].split('*')
    condsamp_pair = [v.split('*') for v in formula.split('|')[1:]]
    condsamp_flat = [c for cs in condsamp_pair for c in cs]
    combinations = list(set(['*'.join(xy)] + ['*'.join(cs) for cs in condsamp_pair]))

    return type('var_', (object,), {'xy': xy, 'condsamp_pair': condsamp_pair, 'condsamp_flat': condsamp_flat, 
                                    'uniquevars': list(set(xy + condsamp_flat)), 'combinations': combinations})


def __fcwt__(input, fs, f0, f1, fn, nthreads=1, scaling="log", fast=False, norm=True, Morlet=6.0):
    """
    function: performs Continuous Wavelet Transform
    call: __cwt__()
    Input:
        a: 
    Return:
        b: 
    """

    #check if input is array and not matrix
    if input.ndim > 1:
        raise ValueError("Input must be a vector")

    #check if input is single precision and change to single precision if not
    if input.dtype != 'single':
        input = input.astype('single')

    morl = fcwt.Morlet(Morlet) #use Morlet wavelet with a wavelet-parameter

    #Generate scales

    if scaling == "lin":
        scales = fcwt.Scales(morl,fcwt.FCWT_LINFREQS,fs,f0,f1,fn)
    elif scaling == "log":
        scales = fcwt.Scales(morl,fcwt.FCWT_LOGSCALES,fs,f0,f1,fn)
    else:
        scales = fcwt.Scales(morl,fcwt.FCWT_LOGSCALES,fs,f0,f1,fn)

    _fcwt = fcwt.FCWT(morl, int(nthreads), fast, norm)

    output = np.zeros((fn,input.size), dtype='csingle')
    freqs = np.zeros((fn), dtype='single')
    
    _fcwt.cwt(input,scales,output)
    scales.getFrequencies(freqs)

    return freqs, output


def __icwt__(W, sj, dt, dj, Cd=None, psi=None, wavelet=None):
    """
    function: performs Inverse Continuous Wavelet Transform
    call: __icwt__()
    Input:
        W: (cross-)spectra
        sj: scales
        dt: sampling rate
        dj: frequency resolution
        Cd, psi: wavelet-specific coefficients
        wavelet: mother wavelet (w/ cdelta and psi(0) callables). Ignored if Cd and psi are given.
    Return:
        x: array
    """
    if wavelet is None: wavelet = pycwt.wavelet.Morlet(6)
    if isinstance(wavelet, str): wavelet = __wavemother_str_pycwt__(wavelet)
    if Cd is None: Cd = wavelet.cdelta
    if psi is None: psi = wavelet.psi(0)
        
    a, b = W.shape
    c = sj.size
    if a == c:
        sj_ = (np.ones([b, 1]) * sj).transpose()
    elif b == c:
        sj_ = np.ones([a, 1]) * sj
    
    x = (W.real / (sj_ ** .5)) * ((dj * dt ** .5) / (Cd * psi))
    return x

   
def __dwt__(signal, level=None, wavelet="db6", **kwargs):
    """
    function: performs Discrete Wavelet Transform
    call: __dwt__()
    Input:
        *args: arrays (1D) to be transformed
        level: maximum scale (power of 2)
        wavelet: mother wavelet (comprehensible to pywt)
    Return:
        Ws: list of 2D arrays
    """
    return pywt.wavedec(signal, wavelet, level=level, **kwargs)


def waverec_2d(coeffs, N, wavelet, mode='symmetric'):
    """
    function: reconstructs 2D signal from wavelet coefficients
    call: waverec_2d()
    Input:
        coeffs: list of wavelet coefficients
        N: data length
        wavelet: mother wavelet (comprehensible to pywt)
        mode: wavelet mode (default is 'symmetric')
    Return:
        rec: reconstructed signal
        """
    def reconstruct_level(coeffs, level_to_keep):
        # Make a copy of the coeffs
        coeffs_copy = [np.zeros_like(c) for c in coeffs]

        # Keep only the desired level coefficients
        coeffs_copy[level_to_keep] = coeffs[level_to_keep]

        # Reconstruct signal
        return pywt.waverec(coeffs_copy, wavelet=wavelet, mode=mode)

    # Example: Reconstruct each level
    reconstructed_levels = []
    for level in range(len(coeffs)):
        rec = reconstruct_level(coeffs, level)
        reconstructed_levels.append(rec[:N])  # Adjust length if needed

    # reconstructed_levels[0] = Approximation (lowest frequency)
    # reconstructed_levels[1], [2], etc. = Details (higher frequencies)
    return reconstructed_levels


def __idwt__(coef, N, wavelet='db6', mode='symmetric'):
    """
    function: performs Inverse Discrete Wavelet Transform
    call: __idwt__()
    Input:
        *args: 2D arrays containing wavelet coefficient
        N: data lenght
        wavelet: mother wavelet (comprehensible to pywt)
    Return:
        Ys: list of 2D arrays
    """
    reconstructed_signal = waverec_2d(coef, N, wavelet, mode=mode)
    reconstructed_signal = np.array(reconstructed_signal[::-1])
    return reconstructed_signal


def prepare_signal(signal, nan_tolerance=0.3, identifier='0000'):
    signal = np.array(signal)
    signan = np.isnan(signal)
    N = len(signal)
    Nnan = np.sum(signan)
    if Nnan:
        if (nan_tolerance > 1 and Nnan > nan_tolerance) or (Nnan > nan_tolerance * N):
            logger.warning(
                f"UserWarning ({identifier}): Too much nans ({np.sum(signan)}, {np.round(100*np.sum(signan)/len(signal), 1)}%).")
    if Nnan and Nnan < N:
        signal = np.interp(np.linspace(0, 1, N),
                            np.linspace(0, 1, N)[
            signan == False],
            signal[signan == False])
    return type('var_', (object,), {'signal': signal, 'signan': signan, 'N': N, 'Nnan': Nnan})


def cone_of_influence(n0, dt, wavelet=None):
    """ function: calculates the cone of influence.
    Uses triangualr Bartlett window with non-zero end-points.
    call: cone_of_influence()
    Input:
        n0: number of points in the signal
        dt: sampling rate (in seconds)
        wavelet: mother wavelet (w/ cdelta and psi(0) callables). If None, uses Morlet.
    Return:
        coi: array with the cone of influence in Fourier periods    
    """

    if wavelet is None: wavelet = pycwt.wavelet.Morlet(6)
    if isinstance(wavelet, str): wavelet = __wavemother_str_pycwt__(wavelet)

    coi = (n0 / 2 - np.abs(np.arange(0, n0) - (n0 - 1) / 2))
    coi = wavelet.flambda() * wavelet.coi() * dt * coi    
    return coi


def cone_of_influence_mask(sj, **kwargs):
    """
    function: calculates the cone of influence for a given wavelet transform
    call: cone_of_influence_mask()
    Input:
        sj: scales
        kwargs: keyword arguments for wavelet transform (e.g., dt, wavelet)
    Return:
        coi: array with True if inside cone of influence, False otherwise
    """
    
    coi = np.array([[s >= (1 / c)
                   for c in cone_of_influence(**kwargs)] for s in sj])

    return coi


def universal_wt(signal, method='dwt', fs=20, f0=1/(3*60*60), f1=10, fn=180, 
                 dj=1/12, wt=True, iwt=True, coi=True, **kwargs):
    """
    function: performs Continuous Wavelet Transform
    call: universal_wt()
    Input:
        signal: 1D array
        method: 'dwt', 'cwt', 'fcwt' (cwt but uses fast algorithm)
        fs: sampling rate (Hz)
        f0: highest scale (becomes level for DWT)
        f1: lowest scale (2x sampling rate)
        fn: number of scales (only used for CWT)
        dj: frequency resolution (only used for CWT)
        iwt: . Default is True
        **kwargs: keyword arguments sent to wavelet transform and inverse functions 
    Return:
        wave: 2D array
        sj: scales 
    """
    assert method in [
        'dwt', 'cwt', 'fcwt'], "Method not found. Available methods are: dwt, cwt, fcwt"

    if method == 'fcwt':
        if fcwt is not None:
            pass
        else:
            logger.warning(
                'UserWarning: Fast continuous wavelet transform (fcwt) not found. Running slow version.')
            method = 'cwt'

    if method == 'cwt':
        if pycwt is not None:
            pass
        else:
            logger.warning(
                'UserWarning: Continuous wavelet transform (cwt) not found. Running discrete version.')
            method = 'dwt'

    if wt:
        # Wavelet Transform
        if method == 'fcwt':
            """Run Continuous Wavelet Transform, using fast algorithm"""
            _l, wave = __fcwt__(signal, fs, f0, f1, fn, **kwargs)
            sj = np.log2(fs/_l)
            sj = list(sj)

        elif method == 'cwt':
            """Run Continuous Wavelet Transform"""
            wave, sj, _, _, _, _ = pycwt.cwt(
                signal, dt=1/fs, s0=2/fs, dj=dj, J=fn-1, **kwargs)
            sj = np.log2(sj*fs)
            sj = list(sj)

        elif method == "dwt":
            """Run Discrete Wavelet Transform"""
            lvl = kwargs.pop('level', int(np.ceil(np.log2(fs/f0))))
            # _l if s0*2^j; fs*2**(-_l) if Hz; (1/fs)*2**_l if sec.
            sj = [_l for _l in np.arange(1, lvl+2, 1)]
            wave = __dwt__(signal, level=lvl, **kwargs)
    else:
        wave = signal

    if iwt:
        # Inverse Wavelet Transform
        if method == 'fcwt':
            sj = kwargs.pop('sj', None)
            wave = __icwt__(wave, sj=sj, dt=fs, dj=dj, **kwargs, 
                        wavelet=pycwt.wavelet.Morlet(6))
        
        elif method == 'cwt':
            sj = kwargs.pop('sj', None)
            wave = __icwt__(wave, sj=sj, dt=fs**-1, dj=dj, **kwargs)
        
        elif method== "dwt":
            N = np.array(signal).shape[-1]
            wave = __idwt__(wave, N=N, **kwargs).real

    if coi:
        # Cone of Influence (COI)
        if method == "dwt":
            coneoi = np.ones_like(wave)
        else:
            coneoi = cone_of_influence_mask(sj=sj, dt=1/fs,
                                        n0=len(signal), 
                                        #wavelet=kwargs.get('mother_wavelet', '')
                                        )
    else:
        coneoi = None
    
    return type('var_', (object,), {
        'wave': wave, 'sj': sj, 'coi': coneoi, 'method': method, 'fs': fs, 'f0': f0, 'f1': f1, 'fn': fn})



# standard modules
import os
import sys
import re
import logging
import datetime
import argparse
# 3rd party modules
import yaml
# Project modules
from . import _core as hc24
from ._extra import eddypro_tools as eddypro
from ._core import handlers


logger = logging.getLogger(__name__)


def set_cwd(workingdir=None):
    """
    Set the current working directory to the specified path.
    
    Parameters:
    - workingdir: Path to the directory to set as current working directory.
    """
    if workingdir is not None:
        os.chdir(workingdir)
        print(f"Current working directory set to: {os.getcwd()}")
        logger.info(f"Current working directory set to: {os.getcwd()}")
    else:
        logger.debug("No working directory specified, using current directory.")


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('-ep', '--eddypro',   type=str,
                        help='Path to EddyPro setup file')
    parser.add_argument('-m', '--metadata',   type=str,
                        help='Path to EddyPro metadata file')
    parser.add_argument('-s', '--site_name',   type=str,
                        help='Site name for the processing')
    parser.add_argument('-i', '--input_path',  type=str,
                        help='Path to the input data folder')
    parser.add_argument('-o', '--output_folderpath', type=str,
                        help='Path to the output folder where results will be saved')
    parser.add_argument('-d', '--datetimerange', type=str,
                        help='Date and time range for processing, format: YYYYMMDD-HHMM-YYYYMMDD-HHMM')
    parser.add_argument('-af', '--acquisition_frequency', type=int, default=20,
                        help='Acquisition frequency in Hz (default: 20 Hz)')
    parser.add_argument('-fd', '--fileduration', type=int, default=30,
                        help='File duration in minutes (default: 30 minutes)')
    parser.add_argument('-ip', '--integration_period', type=int, default=None,
                        help='Integration period in seconds (default: None)')
    parser.add_argument('-v', '--variables_available', type=str, nargs='+',
                        help='List of available variables (default: ["u", "v", "w", "ts", "co2", "h2o"])')
    # parser.add_argument('-dk', '--despike', type=int)  # , nargs=1)
    # parser.add_argument('-dn', '--denoise', type=int)  # , nargs=1)
    # parser.add_argument('-db', '--deadband', type=str, nargs='+')
    parser.add_argument('-cov', '--covariance', type=str, nargs='+',
                        help='List of covariances to compute (default: None)')
    parser.add_argument('--method', type=str, default='dwt', choices=['cov', 'dwt', 'cwt'],
                        help='Method to use for wavelet processing (default: "dwt")')
    parser.add_argument('--wave_mother', type=str, default='db6',
                        help='Mother wavelet to use for wavelet processing (default: "db6")')
    parser.add_argument('--run', type=int, default=1,
                        help='Run the wavelet processing (default: 1)')
    parser.add_argument('--concat', type=int, default=1,
                        help='Concatenate results into a single file (default: 1)')
    parser.add_argument('--partition', type=int, default=1,
                        help='Run partitioning on the results (default: 1)')
    parser.add_argument('--processing_time_duration', type=str, default='1D',
                        help='Processing time duration for the wavelet processing (default: "1D")')
    # parser.add_argument('--preaverage', type=str, default=None)
    parser.add_argument('-cwd', '-workingdir', type=str, default=None,
                        help='Set the current working directory (default: None)')
    parser.add_argument('--log_level', type=str, default='INFO', choices=[
                        'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                        help='Set the logging level')

    args = parser.parse_args()
    return vars(args)


def prepare_args(args):
    # Set the current working directory if specified
    if args.get('cwd'):
        set_cwd(os.chdir(args['cwd']))

    # Set logging level
    log_level = getattr(logging, args.pop(
        'log_level', 'INFO').upper(), logging.INFO)
    args['logging_kwargs'] = {'level': log_level}

    # Extract and pop run flags
    run = args.pop('run', 1)
    concat = args.pop('concat', 1)
    partition = args.pop('partition', 1)
    ep_setup = args.pop('eddypro')
    ep_meta = args.pop('metadata')

    # Retrieve eddypro setup
    # exta = eddypro.extract_info_from_eddypro_setup(ep_setup, ep_meta)
    # args = eddypro.update_args_with_extracted_info(args, exta)

    # Default integration period
    logger.debug(
        f"Integration period {args['integration_period']}, setting to {args['fileduration']} minutes.")
    if args['integration_period'] is None:
        args['integration_period'] = args['fileduration'] * 60

    # Sanitize site name
    args['site_name'] = args['site_name'].replace('/', '_').replace('\\', '_')

    # Convert paths
    for path in ['input_path', 'output_folderpath', 'eddypro', 'metadata']:
        if args.get(path):
            args[path] = os.path.abspath(args[path])

    # Return args and control flags
    return args, run, concat, partition, ep_setup


def validate_args(args):
    required = ['site_name', 'input_path', 'output_folderpath',
                'datetimerange', 'acquisition_frequency', 'fileduration']
    missing = [f'`{k}`' for k in required if args.get(k, None) is None]
    if missing:
        raise ValueError(f'Missing argument(s): {", ".join(missing)}')


def log_args(args):
    # print('\n'.join(
    #     [f'{k}:\t{v[:5] + "~" + v[-25:] if isinstance(v, str) and len(v) > 30 else v}' for k, v in args.items()]), end='\n\n')

    print('Start run with:')
    for k, v in args.items():
        if isinstance(v, str) and len(v) > 30:
            v = f"{v[:5]}~{v[-25:]}"
        print(f"{k}:\t{v}")
    print()


# consider local folder + "/wavelet_flux" as default
# 

def __custom_params__(unknown_args):
    def convert_to_number(value):
        try:
            return int(value)
        except ValueError:
            try:
                return float(value)
            except ValueError:
                return value

    # Print or process unknown arguments
    custom_params = {}
    i = 0
    while i < len(unknown_args):
        logger.debug(f"Custom/unknown arguments: {unknown_args}")
        # You can store or process these as needed, e.g.:
        arg = unknown_args[i]
        if arg.startswith("--"):
            key = arg[2:]
            # Check if the next argument is a value (not starting with '--')
            if i + 1 < len(unknown_args) and not unknown_args[i + 1].startswith("--"):
                custom_params[key] = convert_to_number(unknown_args[i + 1])
                i += 1  # Skip the next argument as it's the value
            else:
                custom_params[key] = True  # Flag argument
        else:
            # Handle values for the previous key if needed
            custom_params[key].append(convert_to_number(arg))
            pass
        i += 1
    return custom_params


def eddypro_run():
    # waveletEC-eddypro_run -p "input/EP/FR-Gri_sample.eddypro" -o "output/wavelet_flux/" -d "20220513T0000-20220516T0000" --processing_time_duration 3H
    parser = argparse.ArgumentParser(description="Run wavelet-based edy covariance workflows.")     
    parser.add_argument('-p', '--path',   type=str,
                        help='Path to EddyPro setup file')
    parser.add_argument('-f', '--folder', type=str,
                        help='Path to the output folder where results will be saved')
    parser.add_argument('-d', '--datetimerange', type=str,
                        help='Date and time range for processing, format: YYYYMMDDTHHMM-YYYYMMDDTHHMM') 
    parser.add_argument('-cov', '--covariance', type=str, nargs='+',
                        help='List of covariances to compute (e.g.: "w*co2", default: None)')
    parser.add_argument("--verbosity", default="INFO", choices=[
                        'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                        help="Logging level (e.g., DEBUG, INFO, WARNING)")
    # Parse known arguments and capture the rest
    args, unknown_args = parser.parse_known_args()

    # Validate logging level
    valid_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    if args.verbosity.upper() not in valid_levels:
        logger.error(
            f"Invalid verbosity level. Choose from: {valid_levels}")
        args.verbosity = "0"

    try:
        custom_params = __custom_params__(unknown_args)
    except UnboundLocalError:
        raise(SyntaxError, 'Check command. Possibly kwargs error, kwargs must be passed as `--key value`.')

    logging.basicConfig(level=args.verbosity.upper(),
                        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    passed_args = {"path": args.path,
                   "output_folderpath": args.folder,
                   "datetimerange": args.datetimerange,
                   "covariance": args.covariance,
                   **custom_params}
    log_args(vars(args))
    handlers.run_from_eddypro(**passed_args)


def integrate():
    parser = argparse.ArgumentParser(
        description="Run wavelet-based edy covariance workflows.")
    parser.add_argument('-f', '--folder', type=str,
                        help='Path to the output folder where results will be saved')
    parser.add_argument('-d', '--dst_path', type=str,
                        help='Path to the output file where results will be saved')    
    parser.add_argument('-ip', '--integration_period', type=int, default=None,
                        help='Integration period in seconds (default: None)')
    parser.add_argument("--verbosity", default="INFO", choices=[
                        'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                        help="Logging level (e.g., DEBUG, INFO, WARNING)")
    # Parse known arguments and capture the rest
    args, unknown_args = parser.parse_known_args()

    args.dst_path = args.dst_path or os.path.join(
        args.folder, '00000_full_cospectra.csv')

    # Validate logging level
    valid_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    if args.verbosity.upper() not in valid_levels:
        logger.error(
            f"Invalid verbosity level. Choose from: {valid_levels}")
        args.verbosity = "0"

    custom_params = __custom_params__(unknown_args)

    logging.basicConfig(level=args.verbosity.upper(),
                        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    passed_args = {"root": os.path.join(args.folder, 'wavelet_full_cospectra'),
                   "f0": 1/args.integration_period,
                   "dst_path": args.dst_path,
                   **custom_params}
    log_args(vars(args))
    handlers.integrate_cospectra_from_file(**passed_args)
    
def partition():
    parser = argparse.ArgumentParser(
        description="Run wavelet-based edy covariance workflows.")
    parser.add_argument('-f', '--folder', type=str,
                        help='Path to the output folder where results will be saved')
    parser.add_argument("--verbosity", default="INFO", choices=[
                        'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                        help="Logging level (e.g., DEBUG, INFO, WARNING)")
    # Parse known arguments and capture the rest
    args, unknown_args = parser.parse_known_args()

    # Validate logging level
    valid_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    if args.verbosity.upper() not in valid_levels:
        logger.error(
            f"Invalid verbosity level. Choose from: {valid_levels}")
        args.verbosity = "0"

    custom_params = __custom_params__(unknown_args)

    logging.basicConfig(level=args.verbosity.upper(),
                        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    passed_args = {"folder": args.folder,
                   **custom_params}
    log_args(vars(args))
    handlers.condition_sampling_partition(**passed_args)


def exec():
    pass

def main(**args):
    # waveleEC --setup "path/to/.eddypro"
    args, run, concat, partition, ep_setup = prepare_args(args)
    validate_args(args)
    log_args(args)

    if args['method'] == 'cov':
        concat = partition = False

    if run:
        handlers.run_from_eddypro(ep_setup, **args)
    # if concat:
    #     integrate_full_spectra_into_file(**args)
    if partition:
        handlers.condition_sampling_partition(**args)


if __name__ == '__main__':
    args = parse_args()
    main(**args)



# standard modules
import os
import sys
import re
import logging
import time
import datetime
import pandas as pd
import numpy as np
# Project modules
import yaml
from waveletec._core import commons as hc24
from waveletec._core import wavelet_functions as wavelet_functions

logger = logging.getLogger('cond_ec')

def run_cec(ymd, raw_kwargs, output_path, 
            w='w', c='co2', q='h2o', averaging=30,
            processing_time_duration="1D", verbosity=1):
    """
    Zahn et al. (2022)
    """
    if isinstance(averaging, (list, tuple)): averaging = averaging[-1]

    info_t_start = time.time()
    logger.info('Entered wavelet code (run_cec).')
    
    suffix = raw_kwargs['suffix'] if 'suffix' in raw_kwargs.keys() else ''

    _, _, _f = ymd
    ymd = hc24.list_time_in_period(*ymd, processing_time_duration, include='both')
            
    logger.info(f'Start date loop at {round(time.time() - info_t_start)} s (run_cec).')
    
    # Skip two line
    prev_print = '\n'
    for ymd_i, yl in enumerate(ymd):
        info_t_startdateloop = time.time()

        date = re.sub('[-: ]', '', str(yl[0]))
        if processing_time_duration.endswith("D"): date = date[:8]
        if processing_time_duration.endswith("H") or processing_time_duration.endswith("Min"): date = date[:12]
        
        print(prev_print, date, 'reading', ' '*10, sep=' ', end='\n')
        prev_print = '\x1B[1A\r'
        
        # load files
        # data = get_rawdata.open_flux(lookup=yl, **raw_kwargs).data
        info_t_startloaddata = time.time()
        data = hc24.loaddatawithbuffer(
            yl, d1=None, freq=_f, buffer=0, f_freq=_f, **raw_kwargs)
        if data.empty:
            if verbosity>1: logger.warning("UserWarning: No file was found ({}, path: {}).".format(date, raw_kwargs.get('path', 'default')))
            continue
        logger.info(f'\tLoading data took {round(time.time() - info_t_startloaddata)} s (run_cec).')
        
        # ensure time is time
        data.TIMESTAMP_ns = pd.to_datetime(data.TIMESTAMP)
        data.TIMESTAMP = data.TIMESTAMP_ns.dt.floor(_f)
        
        try:
            for v in [w, c, q]:
                data[v+'_'] = data.groupby('TIMESTAMP')[v].transform(np.nanmean)
            
            σcec = wavelet_functions.conditional_sampling((data[w]-data[f'{w}_'])*(data[c]-data[f'{c}_']),
                                        *[(data[w]-data[f'{w}_']), (data[c]-data[f'{c}_']), (data[q]-data[f'{q}_'])],
                                        names=['w', 'c', 'q'],
                                        label={1: "+", -1: "-"})
            data['wc'] = (data['w']-data['w_'])*(data['co2']-data['co2_'])
            data['fR'] = σcec['w+c+q+']
            data['fP'] = σcec['w+c-q+']

            σcec = wavelet_functions.conditional_sampling((data[w]-data[f'{w}_'])*(data[q]-data[f'{q}_']),
                                        *[(data[w]-data[f'{w}_']), (data[c]-data[f'{c}_']), (data[q]-data[f'{q}_'])],
                                        names=['w', 'c', 'q'],
                                        label={1: "+", -1: "-"})
            data['wq'] = (data['w']-data['w_'])*(data['h2o']-data['h2o_'])
            data['fE'] = σcec['w+c+q+']
            data['fT'] = σcec['w+c-q+']

            data = data[['TIMESTAMP', w, c, q, 'wc', 'wq', 'fR', 'fP', 'fE', 'fT']].groupby('TIMESTAMP').agg(np.nanmean).reset_index(drop=False)
            data['rFc'] = data['fR'] / data['fP']
            data['Rcec'] = data['wc'] / (1 + ( 1 / data['rFc'] ))
            data['Pcec'] = data['wc'] / (1 + data['rFc'] )
            data['rET'] = data['fE'] / data['fT']
            data['Ecec'] = data['wq'] / (1 + ( 1 / data['rET'] ))
            data['Tcec'] = data['wq'] / (1 + data['rET'] )
            
            __date__ = data['TIMESTAMP'].dt.floor(processing_time_duration).unique()[0]
            dst_path = output_path.format(suffix + "_cec", pd.to_datetime(__date__).strftime('%Y%m%d%H%M'))
            data.to_file(dst_path, index=False)
        except Exception as e:
            logger.critical(str(e))
            print(str(e))
        
        prev_print = '\x1B[2A\r' + f' {date} {len(yl)} files {int(100*ymd_i/len(ymd))} % ({time.strftime("%d.%m.%y %H:%M:%S")})' + '\n'
            
        logger.info(f'\tFinish {date} took {round(time.time() - info_t_startdateloop)} s, yielded {len(yl)} files (run_cec). Progress: {len(yl)} {int(100*ymd_i/len(ymd))} %')
    
    logger.info(f'Finish conditional EC (Zahn et al., 2022) at {round(time.time() - info_t_start)} s (run_cec).')
    return


def main(sitename, inputpath, outputpath, datetimerange, acquisition_frequency=20, fileduration=30, processduration='1D',
         **kwargs):
    local_args = locals()

    logname = str(os.path.join(outputpath, f"log/current_{datetime.datetime.now().strftime('%y%m%dT%H%M%S')}.log"))
    hc24.mkdirs(logname)
    #with open(logname, "w+"): pass
    logging.basicConfig(filename=logname,
                        filemode='a',
                        format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S',
                        level=logging.DEBUG, 
                        force=True)

    logging.captureWarnings(True)
    logging.info("STARTING THE RUN")

    # Create setup
    configure = hc24.structuredData()

    configure.processing_time_duration = processduration

    # Select output file path
    configure.output_path = str(os.path.join(outputpath, 'processing', str(sitename)+'_CEC{}_{}.csv'))

    # Save args for run
    hc24.mkdirs(configure.output_path)
    with open(os.path.join(os.path.dirname(os.path.dirname(configure.output_path)), f'cec_{datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")}.setup'), 'w+') as stp:
        yaml.safe_dump(local_args, stp)

    # Select raw input file path
    # e.g.: "<PROJECT FOLDER>/eddypro_output/eddypro_raw_datasets/level_6/"
    if isinstance(inputpath, dict):
        configure.raw_kwargs = {'path': {k: str(os.path.join(v, 'eddypro_raw_datasets/level_6')) for k, v in inputpath.items()}}
    else: configure.raw_kwargs = {'path': str(os.path.join(inputpath, 'eddypro_raw_datasets/level_6'))}

    # Select period of interest
    # [START_DATE, END_DATE, FILE_FREQUENCY]
    configure.ymd = [datetimerange.split('-')[0], datetimerange.split('-')[1], f'{fileduration}min']

    # Averaging (and integrating) time
    configure.averaging = [fileduration]

    # Select dt
    configure.raw_kwargs.update({'fkwargs': {'dt': 1/acquisition_frequency},
                                 'fmt': {'co': '4th'}})

    # RUN WAVELET FLUX PROCESSING
    run_cec(**vars(configure), verbosity=5)


def __concat__(sitename, outputpath, **kwargs):
    # CONCAT INTO SINGLE FILE
    dst_path = os.path.join(outputpath, str(sitename)+f'_CEC_partition.csv')

    root = os.path.join(outputpath, 'processing')
    print(os.listdir(root))
    saved_files = {}
    for name in os.listdir(root):
        dateparts = re.findall('_([0-9]{12}).csv$', name, flags=re.IGNORECASE)
        if len(dateparts) == 1:
            saved_files[dateparts[0]] = os.path.join(root, name)

    data = pd.concat([pd.read_csv(v, sep=',') for k, v in saved_files.items()])

    if dst_path: data.to_file(dst_path, index=False)
    else: return data
    return


def handle_eddypro_setup(**args):
    if args['eddypro']:
        eddypro_setup = hc24.read_eddypro_metadata_file(args['eddypro'])
        if args['sitename'] is None: args['sitename'] = eddypro_setup['Project']['project_title']
        if args['inputpath'] is None: args['inputpath'] = eddypro_setup['Project']['out_path']
        if args['outputpath'] is None: args['outputpath'] = eddypro_setup['Project']['out_path'] + '/wavelet_flux'
        if args['datetimerange'] is None: args['datetimerange'] = eddypro_setup['Project']['pr_start_date'].replace('-', '') + eddypro_setup['Project']['pr_start_time'].replace(':', '') + '-' + \
            eddypro_setup['Project']['pr_end_date'].replace('-', '') + eddypro_setup['Project']['pr_end_time'].replace(':', '')
        if args['fileduration'] is None: args['fileduration'] = int(eddypro_setup['RawProcess_Settings']['avrg_len'])
    
        if args['metadata'] is None: 
            if eddypro_setup['Project']['proj_file']: args['metadata'] = eddypro_setup['Project']['proj_file']
            else: args['metadata'] = args['eddypro'].rsplit('.', 1)[0] + '.metadata'
    
    if args['metadata']:
        eddypro_metad = hc24.read_eddypro_metadata_file(args['metadata'])
        if args['acquisition_frequency'] is None: args['acquisition_frequency'] = int(float(eddypro_metad['Timing']['acquisition_frequency']))
        #if args['fileduration'] is None: args['fileduration'] = int(eddypro_metad['Timing']['file_duration'])
    
    return args


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('-ep', '--eddypro',   type=str)
    parser.add_argument('-m', '--metadata',   type=str)
    parser.add_argument('-s', '--sitename',   type=str)
    parser.add_argument('-i', '--inputpath',  type=str)
    parser.add_argument('-o', '--outputpath', type=str)
    parser.add_argument('-d', '--datetimerange', type=str)
    parser.add_argument('-af', '--acquisition_frequency', type=int)
    parser.add_argument('-fd', '--fileduration', type=int)
    parser.add_argument('--run', type=int, default=1)
    parser.add_argument('--concat', type=int, default=1)
    parser.add_argument('--processduration', type=str, default='1D')
    args = parser.parse_args()
    args = handle_eddypro_setup(**vars(args))
    
    args['sitename'] = args['sitename'].replace('/', '_').replace('\\', '_')
    
    args.pop('eddypro')
    args.pop('metadata')
    run = args.pop('run')
    concat = args.pop('concat')

    print('Start run w/')
    # replace os.get_cwd() for '' if str
    print('\n'.join([f'{k}:\t{v[:5] + "~" + v[-25:] if isinstance(v, str) and len(v) > 30 else v}' for k, v in args.items()]), end='\n\n')
    
    # Assert variables have been assigned
    missing_args = [f'`{k}`' for k in ['sitename', 'inputpath', 'outputpath', 'datetimerange', 'acquisition_frequency', 'fileduration'] if args[k] is None]
    assert len(missing_args) == 0, f'Missing argument in: {", ".join(missing_args)}.'

    if run: main(**args)
    if concat: __concat__(**args)



# built-in modules
import re
import os
import warnings
import logging
from functools import reduce

# 3rd party modules
import numpy as np
import pandas as pd

# project modules
from .._core.commons import get_all_sites
from .._core import corrections

logger = logging.getLogger('wvlt.eddypro_tools')

def read_eddypro_metadata_file(filename):
    metadata = {}
    with open(filename, 'r') as file:
        section = None
        for line in file:
            line = line.strip()
            if line.startswith(';') or not line:
                continue  # Skip comments and empty lines
            if line.startswith('[') and line.endswith(']'):
                section = line[1:-1]  # Extract section name
                metadata[section] = {}
            else:
                key, value = line.split('=', 1)
                metadata[section][key.strip()] = value.strip()
    return metadata


def get_eddypro_output(site_name=None, treated=True, **kwargs):
    if site_name is None:
        return get_all_sites(get_eddypro_output, **kwargs)
    mergeorconcat = kwargs.get('mergeorconcat', 'merge')
    folder        = kwargs.get('folder', 'data')
    
    ep_read_params = {'skiprows': [0,2], 'na_values': [-9999, 'NAN']}
    
    ep_path = os.path.join(folder, site_name, 'eddypro_output')

    files = {'FLUX': [], 'FLXNT': [], 'QCQA': [], 'META': []}
    for name in os.listdir(ep_path):
        if name.endswith('.csv'):
            if re.findall('_full_output_', name):
                if name.endswith('_adv.csv'):
                    files['FLUX'] += [pd.read_csv(os.path.join(ep_path, name), **ep_read_params)]
                else:
                    files['FLUX'] += [pd.read_csv(os.path.join(ep_path, name), na_values=[-9999, 'NAN'])]
            elif re.findall('_fluxnet_', name):
                flxnt_last = pd.read_csv(os.path.join(ep_path, name), na_values=[-9999, 'NAN'])
                flxnt_last["date"] = pd.to_datetime(flxnt_last["TIMESTAMP_START"], format='%Y%m%d%H%M').dt.strftime('%Y-%m-%d')
                flxnt_last["time"] = pd.to_datetime(flxnt_last["TIMESTAMP_START"], format='%Y%m%d%H%M').dt.strftime('%H:%M')
                files['FLXNT'] += [flxnt_last]
                del flxnt_last
            elif re.findall('_qc_details_', name):
                if name.endswith('_adv.csv'):
                    files['QCQA'] += [pd.read_csv(os.path.join(ep_path, name), **ep_read_params)]
                else:
                    files['QCQA'] += [pd.read_csv(os.path.join(ep_path, name), na_values=[-9999, 'NAN'])]
            elif re.findall('_metadata_', name):
                files['META'] += [pd.read_csv(os.path.join(ep_path, name), na_values=[-9999, 'NAN'])]
    
    for k in [k for k in list(files.keys()) if not files[k]]:
        del([files[k]])

    for k in files.keys():
        if len(files[k]) == 1:
            files[k] = files[k][0]
        elif mergeorconcat == 'concat':
            files[k] = pd.concat(files[k])
        else:
            files[k] = reduce(lambda left, right: pd.merge(
                left, right, on=["date", 'time'], how="outer", suffixes=('', '_DUP')), files[k])
    
    data = pd.DataFrame(files.pop('FLUX', {}))
    for name, dat in files.items():
        data = pd.merge(data, dat, on=["date", 'time'], how="outer", suffixes=('', f'_{name}'))
    
    data['TIMESTAMP'] = pd.to_datetime(data.date + ' ' + data.time)#.dt.tz_localize('UTC')
    
    if treated:
        # Convert to UTC
        data['TIMESTAMP_UTC'] = data['TIMESTAMP'].dt.tz_localize('UTC')

        # Despike because open path analysers can be noisy
        data['air_molar_volume_despiked'] = corrections.mauder2013(data.air_molar_volume, 5)
        data['Vd'] = (data.air_molar_volume_despiked * data.air_pressure /
                            (data.air_pressure - data.e))
        # Make wet from dry
        data['Va'] = (data.air_molar_volume_despiked * data.air_pressure /
                            (data.air_pressure - data.e))
    return data


def get_eddypro_cospectra(site_name=None, x='natural_frequency', y='f_nat*cospec(w_co2)/cov(w_co2)', folder='data', subfolder='', help=False):
    assert (x is not None and y is not None) or (help)
    if site_name is None:
        return get_all_sites(get_eddypro_cospectra, x=x, y=y, folder=folder, subfolder=subfolder, help=help)
    
    ep_path = os.path.join(folder, site_name, 'output/eddypro_output', subfolder, 'eddypro_binned_cospectra')
    
    if not os.path.exists(ep_path): return None
    files = []
    for name in os.listdir(ep_path):
        if name.endswith('.csv'):
            if re.findall('_binned_cospectra_', name):
                binned_cosp = pd.read_csv(os.path.join(ep_path, name), skiprows=11, na_values=[-9999, 'NAN'])
                if help: 
                    print(binned_cosp.columns)
                    return
                if x not in binned_cosp.columns or y not in binned_cosp.columns: continue
                binned_cosp.dropna(subset=[x], inplace=True)
                binned_cosp = binned_cosp[[x, y]]
                binned_cosp['TIMESTAMP'] = name.split('_')[0]
                binned_cosp = binned_cosp.pivot(index='TIMESTAMP', columns=x, values=y).reset_index(drop=False)
                binned_cosp.columns = [c for c in binned_cosp.columns]
                files += [binned_cosp]
    
    if len(files) == 1:
        data = files[0]
    else:
        data = pd.concat(files)
    
    data['TIMESTAMP'] = pd.to_datetime(data['TIMESTAMP'], format='%Y%m%d-%H%M')
    return data


def extract_info_from_eddypro_setup(eddypro=None, metadata=None):
    """
    Extracts information from EddyPro setup and metadata files and returns a dictionary.

    Parameters:
    - eddypro: Path to the EddyPro setup file.
    - metadata: Path to the metadata file.

    Returns:
    - A dictionary containing extracted information.
    """    
    args = {}

    if eddypro:
        eddypro_setup = read_eddypro_metadata_file(eddypro)

        metadata = metadata or eddypro_setup['Project']['proj_file'] or str(eddypro).rsplit('.', 1)[
            0] + '.metadata'

        args['site_name'] = eddypro_setup['Project']['project_id']
        args['input_path'] = eddypro_setup['Project']['out_path'] + \
            '/eddypro_raw_datasets/level_6'
        args['output_folderpath'] = eddypro_setup['Project']['out_path'] + \
            '/wavelet_flux'
        args['datetimerange'] = (
            eddypro_setup['Project']['pr_start_date'].replace('-', '') +
            eddypro_setup['Project']['pr_start_time'].replace(':', '') + '-' +
            eddypro_setup['Project']['pr_end_date'].replace('-', '') +
            eddypro_setup['Project']['pr_end_time'].replace(':', '')
        )
        args['fileduration'] = int(
            eddypro_setup['RawProcess_Settings']['avrg_len'])

        try:
            gas4_col = eddypro_setup['Project']['col_n2o']
            eddypro_metad = read_eddypro_metadata_file(metadata)
            if (not gas4_col) or (str(gas4_col) != '0'):
                args['gas4_name'] = eddypro_metad['FileDescription'][f'col_{gas4_col}_variable']
        except Exception as e:
            print(f"Error extracting gas4 name: {e}")

    if metadata:
        eddypro_metad = read_eddypro_metadata_file(metadata)
        args['acquisition_frequency'] = int(
            float(eddypro_metad['Timing']['acquisition_frequency']))

    if eddypro:
        args['variables_available'] = ['u', 'v', 'w'] + [k for k in ['co2',
                                                                     'h2o', 'ch4'] if float(eddypro_setup['Project'][f'col_{k}']) > 0]

    if metadata and float(eddypro_setup['Project']['col_n2o']) > 0:
        gas4 = eddypro_metad['FileDescription'][f"col_{eddypro_setup['Project']['col_n2o']}_variable"]
        if gas4:
            args['variables_available'] = args.get(
                'variables_available', []) + [gas4]

    return args


def update_args_with_extracted_info(args, extracted_info):
    """
    Updates the args dictionary with the extracted information.

    Parameters:
    - args: The dictionary to be updated.
    - extracted_info: The dictionary containing extracted information.

    Returns:
    - The updated args dictionary.
    """
    for key, value in extracted_info.items():
        if key not in args or args[key] is None:
            args[key] = value
    return args
